@@ -1,3 +1,4 @@
+import copy
 import math
 import pickle
 
@@ -8,16 +9,19 @@ from datetime import timedelta
 
 import pytest
 
-from tomlkit import inline_table
+from tests.util import assert_is_ppo
+from tests.util import elementary_test
+from tomlkit import api
 from tomlkit import parse
-from tomlkit.api import array
-from tomlkit.api import ws
 from tomlkit.exceptions import NonExistentKey
+from tomlkit.items import Array
 from tomlkit.items import Bool
 from tomlkit.items import Comment
 from tomlkit.items import InlineTable
 from tomlkit.items import Integer
+from tomlkit.items import Item
 from tomlkit.items import KeyType
+from tomlkit.items import Null
 from tomlkit.items import SingleKey as Key
 from tomlkit.items import String
 from tomlkit.items import StringType
@@ -27,6 +31,140 @@ from tomlkit.items import item
 from tomlkit.parser import Parser
 
 
+@pytest.fixture()
+def tz_pst():
+    try:
+        from datetime import timezone
+
+        return timezone(timedelta(hours=-8), "PST")
+    except ImportError:
+        from datetime import tzinfo
+
+        class PST(tzinfo):
+            def utcoffset(self, dt):
+                return timedelta(hours=-8)
+
+            def tzname(self, dt):
+                return "PST"
+
+            def dst(self, dt):
+                return timedelta(0)
+
+        return PST()
+
+
+@pytest.fixture()
+def tz_utc():
+    try:
+        from datetime import timezone
+
+        return timezone.utc
+    except ImportError:
+        from datetime import tzinfo
+
+        class UTC(tzinfo):
+            def utcoffset(self, dt):
+                return timedelta(hours=0)
+
+            def tzname(self, dt):
+                return "UTC"
+
+            def dst(self, dt):
+                return timedelta(0)
+
+        return UTC()
+
+
+def test_item_base_has_no_unwrap():
+    trivia = Trivia(indent="\t", comment_ws=" ", comment="For unit test")
+    item = Item(trivia)
+    try:
+        item.unwrap()
+    except NotImplementedError:
+        pass
+    else:
+        raise AssertionError("`items.Item` should not implement `unwrap`")
+
+
+def test_integer_unwrap():
+    elementary_test(item(666), int)
+
+
+def test_float_unwrap():
+    elementary_test(item(2.78), float)
+
+
+def test_false_unwrap():
+    elementary_test(item(False), bool)
+
+
+def test_true_unwrap():
+    elementary_test(item(True), bool)
+
+
+def test_datetime_unwrap():
+    dt = datetime.utcnow()
+    elementary_test(item(dt), datetime)
+
+
+def test_string_unwrap():
+    elementary_test(item("hello"), str)
+
+
+def test_null_unwrap():
+    n = Null()
+    elementary_test(n, type(None))
+
+
+def test_aot_unwrap():
+    d = item([{"a": "A"}, {"b": "B"}])
+    unwrapped = d.unwrap()
+    assert_is_ppo(unwrapped, list)
+    for du, _ in zip(unwrapped, d):
+        assert_is_ppo(du, dict)
+        for ku in du:
+            vu = du[ku]
+            assert_is_ppo(ku, str)
+            assert_is_ppo(vu, str)
+
+
+def test_time_unwrap():
+    t = time(3, 8, 14)
+    elementary_test(item(t), time)
+
+
+def test_date_unwrap():
+    d = date.today()
+    elementary_test(item(d), date)
+
+
+def test_array_unwrap():
+    trivia = Trivia(indent="\t", comment_ws=" ", comment="For unit test")
+    i = item(666)
+    f = item(2.78)
+    b = item(False)
+    a = Array([i, f, b], trivia)
+    a_unwrapped = a.unwrap()
+    assert_is_ppo(a_unwrapped, list)
+    assert_is_ppo(a_unwrapped[0], int)
+    assert_is_ppo(a_unwrapped[1], float)
+    assert_is_ppo(a_unwrapped[2], bool)
+
+
+def test_abstract_table_unwrap():
+    table = item({"foo": "bar"})
+    super_table = item({"table": table, "baz": "borg"})
+
+    table_unwrapped = super_table.unwrap()
+    sub_table = table_unwrapped["table"]
+    assert_is_ppo(table_unwrapped, dict)
+    assert_is_ppo(sub_table, dict)
+    for ku in sub_table:
+        vu = sub_table[ku]
+        assert_is_ppo(ku, str)
+        assert_is_ppo(vu, str)
+
+
 def test_key_comparison():
     k = Key("foo")
 
@@ -44,26 +182,26 @@ def test_items_can_be_appended_to_and_removed_from_a_table():
     _, table = parser._parse_table()
 
     assert isinstance(table, Table)
-    assert "" == table.as_string()
+    assert table.as_string() == ""
 
     table.append(Key("foo"), String(StringType.SLB, "bar", "bar", Trivia(trail="\n")))
 
-    assert 'foo = "bar"\n' == table.as_string()
+    assert table.as_string() == 'foo = "bar"\n'
 
     table.append(
         Key("baz"),
         Integer(34, Trivia(comment_ws="   ", comment="# Integer", trail=""), "34"),
     )
 
-    assert 'foo = "bar"\nbaz = 34   # Integer' == table.as_string()
+    assert table.as_string() == 'foo = "bar"\nbaz = 34   # Integer'
 
     table.remove(Key("baz"))
 
-    assert 'foo = "bar"\n' == table.as_string()
+    assert table.as_string() == 'foo = "bar"\n'
 
     table.remove(Key("foo"))
 
-    assert "" == table.as_string()
+    assert table.as_string() == ""
 
     with pytest.raises(NonExistentKey):
         table.remove(Key("foo"))
@@ -77,23 +215,23 @@ def test_items_can_be_appended_to_and_removed_from_an_inline_table():
     _, table = parser._parse_item()
 
     assert isinstance(table, InlineTable)
-    assert "{}" == table.as_string()
+    assert table.as_string() == "{}"
 
     table.append(Key("foo"), String(StringType.SLB, "bar", "bar", Trivia(trail="")))
 
-    assert '{foo = "bar"}' == table.as_string()
+    assert table.as_string() == '{foo = "bar"}'
 
     table.append(Key("baz"), Integer(34, Trivia(trail=""), "34"))
 
-    assert '{foo = "bar", baz = 34}' == table.as_string()
+    assert table.as_string() == '{foo = "bar", baz = 34}'
 
     table.remove(Key("baz"))
 
-    assert '{foo = "bar"}' == table.as_string()
+    assert table.as_string() == '{foo = "bar"}'
 
     table.remove(Key("foo"))
 
-    assert "{}" == table.as_string()
+    assert table.as_string() == "{}"
 
     with pytest.raises(NonExistentKey):
         table.remove(Key("foo"))
@@ -201,7 +339,7 @@ def test_array_multiline():
     t = item([])
     t.multiline(True)
 
-    assert "[]" == t.as_string()
+    assert t.as_string() == "[]"
 
 
 def test_array_multiline_modify():
@@ -231,7 +369,7 @@ a = [
 def test_append_to_empty_array():
     doc = parse("x = [ ]")
     doc["x"].append("a")
-    assert doc.as_string() == 'x = [ "a" ]'
+    assert doc.as_string() == 'x = ["a" ]'
     doc = parse("x = [\n]")
     doc["x"].append("a")
     assert doc.as_string() == 'x = [\n    "a"\n]'
@@ -273,12 +411,49 @@ x = [
     2
 ]"""
     )
+    doc["x"].pop(0)
+    assert doc.as_string() == "x = [\n    2\n]"
+
+
+def test_append_to_multiline_array_with_comment():
+    doc = parse(
+        """\
+x = [
+    # Here is a comment
+    1,
+    2
+]
+"""
+    )
+    doc["x"].multiline(True).append(3)
+    assert (
+        doc.as_string()
+        == """\
+x = [
+    # Here is a comment
+    1,
+    2,
+    3,
+]
+"""
+    )
+    assert doc["x"].pop() == 3
+    assert (
+        doc.as_string()
+        == """\
+x = [
+    # Here is a comment
+    1,
+    2,
+]
+"""
+    )
 
 
 def test_append_dict_to_array():
-    doc = parse("x = [ ]")
+    doc = parse("x = []")
     doc["x"].append({"name": "John Doe", "email": "john@doe.com"})
-    expected = 'x = [ {name = "John Doe",email = "john@doe.com"} ]'
+    expected = 'x = [{name = "John Doe",email = "john@doe.com"}]'
     assert doc.as_string() == expected
     # Make sure the produced string is valid
     assert parse(doc.as_string()) == doc
@@ -296,10 +471,11 @@ bar = "baz"
 
 
 def test_array_add_line():
-    t = array()
+    t = api.array()
     t.add_line(1, 2, 3, comment="Line 1")
     t.add_line(4, 5, 6, comment="Line 2")
-    t.add_line(7, ws(","), ws(" "), 8, add_comma=False)
+    t.add_line(7, api.ws(","), api.ws(" "), 8, add_comma=False)
+    t.add_line(comment="Line 4")
     t.add_line(indent="")
     assert len(t) == 8
     assert list(t) == [1, 2, 3, 4, 5, 6, 7, 8]
@@ -308,15 +484,16 @@ def test_array_add_line():
         == """[
     1, 2, 3, # Line 1
     4, 5, 6, # Line 2
-    7, 8
+    7, 8,
+    # Line 4
 ]"""
     )
 
 
 def test_array_add_line_invalid_value():
-    t = array()
+    t = api.array()
     with pytest.raises(ValueError, match="is not allowed"):
-        t.add_line(1, ws(" "))
+        t.add_line(1, api.ws(" "))
     with pytest.raises(ValueError, match="is not allowed"):
         t.add_line(Comment(Trivia("  ", comment="test")))
     assert len(t) == 0
@@ -401,6 +578,13 @@ bar = "baz"
     )
 
 
+def test_add_sum_int_with_float():
+    content = "[table]\nmy_int = 2048.3"
+    doc = parse(content)
+    doc["table"]["my_int"] += 5
+    assert doc["table"]["my_int"] == 2053.3
+
+
 def test_integers_behave_like_ints():
     i = item(34)
 
@@ -441,7 +625,7 @@ def test_floats_behave_like_floats():
     assert doc.as_string() == "float = +35.12"
 
 
-def test_datetimes_behave_like_datetimes():
+def test_datetimes_behave_like_datetimes(tz_utc, tz_pst):
     i = item(datetime(2018, 7, 22, 12, 34, 56))
 
     assert i == datetime(2018, 7, 22, 12, 34, 56)
@@ -455,6 +639,14 @@ def test_datetimes_behave_like_datetimes():
     assert i == datetime(2018, 7, 21, 12, 34, 56)
     assert i.as_string() == "2018-07-21T12:34:56"
 
+    i = i.replace(year=2019, tzinfo=tz_utc)
+    assert i == datetime(2019, 7, 21, 12, 34, 56, tzinfo=tz_utc)
+    assert i.as_string() == "2019-07-21T12:34:56+00:00"
+
+    i = i.astimezone(tz_pst)
+    assert i == datetime(2019, 7, 21, 4, 34, 56, tzinfo=tz_pst)
+    assert i.as_string() == "2019-07-21T04:34:56-08:00"
+
     doc = parse("dt = 2018-07-22T12:34:56-05:00")
     doc["dt"] += timedelta(days=1)
 
@@ -475,6 +667,10 @@ def test_dates_behave_like_dates():
     assert i == date(2018, 7, 21)
     assert i.as_string() == "2018-07-21"
 
+    i = i.replace(year=2019)
+    assert i == datetime(2019, 7, 21)
+    assert i.as_string() == "2019-07-21"
+
     doc = parse("dt = 2018-07-22 # Comment")
     doc["dt"] += timedelta(days=1)
 
@@ -487,6 +683,10 @@ def test_times_behave_like_times():
     assert i == time(12, 34, 56)
     assert i.as_string() == "12:34:56"
 
+    i = i.replace(hour=13)
+    assert i == time(13, 34, 56)
+    assert i.as_string() == "13:34:56"
+
 
 def test_strings_behave_like_strs():
     i = item("foo")
@@ -508,6 +708,14 @@ def test_strings_behave_like_strs():
     assert doc.as_string() == 'str = "foo bar" # Comment'
 
 
+def test_string_add_preserve_escapes():
+    i = api.value('"foo\\"bar"')
+    i += " baz"
+
+    assert i == 'foo"bar baz'
+    assert i.as_string() == '"foo\\"bar baz"'
+
+
 def test_tables_behave_like_dicts():
     t = item({"foo": "bar"})
 
@@ -587,7 +795,7 @@ def test_items_are_pickable():
     s = pickle.dumps(n)
     assert pickle.loads(s).as_string() == 'foo = "bar"\n'
 
-    n = inline_table()
+    n = api.inline_table()
     n["foo"] = "bar"
 
     s = pickle.dumps(n)
@@ -605,7 +813,7 @@ def test_items_are_pickable():
 
 
 def test_trim_comments_when_building_inline_table():
-    table = inline_table()
+    table = api.inline_table()
     row = parse('foo = "bar"  # Comment')
     table.update(row)
     assert table.as_string() == '{foo = "bar"}'
@@ -613,28 +821,28 @@ def test_trim_comments_when_building_inline_table():
     value.comment("Another comment")
     table.append("baz", value)
     assert "# Another comment" not in table.as_string()
-    assert '{foo = "bar", baz = "foobaz"}' == table.as_string()
+    assert table.as_string() == '{foo = "bar", baz = "foobaz"}'
 
 
 def test_deleting_inline_table_elemeent_does_not_leave_trailing_separator():
-    table = inline_table()
+    table = api.inline_table()
     table["foo"] = "bar"
     table["baz"] = "boom"
 
-    assert '{foo = "bar", baz = "boom"}' == table.as_string()
+    assert table.as_string() == '{foo = "bar", baz = "boom"}'
 
     del table["baz"]
 
-    assert '{foo = "bar"}' == table.as_string()
+    assert table.as_string() == '{foo = "bar"}'
 
-    table = inline_table()
+    table = api.inline_table()
     table["foo"] = "bar"
 
     del table["foo"]
 
     table["baz"] = "boom"
 
-    assert '{baz = "boom"}' == table.as_string()
+    assert table.as_string() == '{baz = "boom"}'
 
 
 def test_booleans_comparison():
@@ -663,3 +871,27 @@ def test_table_copy():
     table["foo"] = "baz"
     assert table_copy["foo"] == "bar"
     assert table_copy.as_string() == 'foo = "bar"\n'
+
+
+def test_copy_copy():
+    result = parse(
+        """
+    [tool.poetry]
+    classifiers = [
+    # comment
+        "a",
+        "b",
+    ]
+    """
+    )
+    classifiers = result["tool"]["poetry"]["classifiers"]
+    new = copy.copy(classifiers)
+    assert new == classifiers
+
+
+@pytest.mark.parametrize(
+    "key_str,escaped",
+    [("\\", '"\\\\"'), ('"', '"\\""'), ("\t", '"\\t"'), ("\x10", '"\\u0010"')],
+)
+def test_escape_key(key_str, escaped):
+    assert api.key(key_str).as_string() == escaped
